=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-10-07 06:13:14 +0000
+++ bzrlib/errors.py	2005-10-08 00:32:39 +0000
@@ -197,7 +197,7 @@
         BzrError.__init__(self, "Stores for branch %s are not listable" % br)
 
 
-from bzrlib.weave import WeaveError
+from bzrlib.weave import WeaveError, WeaveParentMismatch
 
 class TransportError(BzrError):
     """All errors thrown by Transport implementations should derive

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-10-07 06:13:14 +0000
+++ bzrlib/fetch.py	2005-10-08 00:32:39 +0000
@@ -17,13 +17,14 @@
 import os
 from cStringIO import StringIO
 
-import bzrlib.errors
+import bzrlib
+import bzrlib.errors as errors
+from bzrlib.errors import InstallFailed, NoSuchRevision, WeaveError
 from bzrlib.trace import mutter, note, warning
 from bzrlib.branch import Branch
 from bzrlib.progress import ProgressBar
 from bzrlib.xml5 import serializer_v5
 from bzrlib.osutils import sha_string, split_lines
-from bzrlib.errors import InstallFailed, NoSuchRevision, WeaveError
 
 """Copying of history from one branch to another.
 
@@ -218,11 +219,14 @@
         if file_id in self.copied_file_ids:
             mutter('file {%s} already copied', file_id)
             return
-        from_weave = self.from_weaves.get_weave(file_id, 
+        from_weave = self.from_weaves.get_weave(file_id,
             self.from_branch.get_transaction())
         to_weave = self.to_weaves.get_weave_or_empty(file_id,
             self.to_branch.get_transaction())
-        to_weave.join(from_weave)
+        try:
+            to_weave.join(from_weave)
+        except errors.WeaveParentMismatch:
+            to_weave.reweave(from_weave)
         self.to_weaves.put_weave(file_id, to_weave,
             self.to_branch.get_transaction())
         self.count_weaves += 1

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2005-10-07 06:13:14 +0000
+++ bzrlib/inventory.py	2005-10-08 00:32:39 +0000
@@ -166,6 +166,14 @@
                 ie = inv[self.file_id]
                 assert ie.file_id == self.file_id
                 if ie.revision in heads:
+                    # fixup logic, there was a bug in revision updates.
+                    # with x bit support.
+                    try:
+                        if heads[ie.revision].executable != ie.executable:
+                            heads[ie.revision].executable = False
+                            ie.executable = False
+                    except AttributeError:
+                        pass
                     assert heads[ie.revision] == ie
                 else:
                     # may want to add it.

=== modified file 'bzrlib/selftest/test_weave.py'
--- bzrlib/selftest/test_weave.py	2005-10-06 10:53:12 +0000
+++ bzrlib/selftest/test_weave.py	2005-10-08 00:03:38 +0000
@@ -18,14 +18,15 @@
 
 
 # TODO: tests regarding version names
-
-
+# TODO: rbc 20050108 test that join does not leave an inconsistent weave 
+#       if it fails.
 
 """test suite for weave algorithm"""
 
 from pprint import pformat
 
-from bzrlib.weave import Weave, WeaveFormatError, WeaveError
+import bzrlib.errors as errors
+from bzrlib.weave import Weave, WeaveFormatError, WeaveError, reweave
 from bzrlib.weavefile import write_weave, read_weave
 from bzrlib.selftest import TestCase
 from bzrlib.osutils import sha_string
@@ -873,7 +874,24 @@
         eq(sorted(wa.iter_names()), ['v1', 'v2', 'v3', 'x1',])
         eq(wa.get_text('x1'), 'line from x1\n')
 
-    def test_join_with_ghosts(self):
+    def test_reweave_with_empty(self):
+        wb = Weave()
+        wr = reweave(self.weave1, wb)
+        eq = self.assertEquals
+        eq(sorted(wr.iter_names()), ['v1', 'v2', 'v3'])
+        eq(wr.get_lines('v3'), ['hello\n', 'cruel\n', 'world\n'])
+        self.weave1.reweave(wb)
+        self.assertEquals(wr, self.weave1)
+
+    def test_join_with_ghosts_raises_parent_mismatch(self):
+        wa = self.weave1.copy()
+        wb = Weave()
+        wb.add('x1', [], ['line from x1\n'])
+        wb.add('v1', [], ['hello\n'])
+        wb.add('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
+        self.assertRaises(errors.WeaveParentMismatch, wa.join, wb)
+
+    def test_reweave_with_ghosts(self):
         """Join that inserts parents of an existing revision.
 
         This can happen when merging from another branch who
@@ -881,16 +899,19 @@
         this test the second weave knows of an additional parent of 
         v2.  Any revisions which are in common still have to have the 
         same text."""
-        return ###############################
         wa = self.weave1.copy()
         wb = Weave()
         wb.add('x1', [], ['line from x1\n'])
         wb.add('v1', [], ['hello\n'])
         wb.add('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        wa.join(wb)
+        wc = reweave(wa, wb)
         eq = self.assertEquals
-        eq(sorted(wa.iter_names()), ['v1', 'v2', 'v3', 'x1',])
-        eq(wa.get_text('x1'), 'line from x1\n')
+        eq(sorted(wc.iter_names()), ['v1', 'v2', 'v3', 'x1',])
+        eq(wc.get_text('x1'), 'line from x1\n')
+        eq(wc.get_lines('v2'), ['hello\n', 'world\n'])
+        eq(wc.parent_names('v2'), ['v1', 'x1'])
+        self.weave1.reweave(wb)
+        self.assertEquals(wc, self.weave1)
 
 
 if __name__ == '__main__':

=== modified file 'bzrlib/selftest/testfetch.py'
--- bzrlib/selftest/testfetch.py	2005-10-06 06:31:33 +0000
+++ bzrlib/selftest/testfetch.py	2005-10-08 00:03:38 +0000
@@ -96,6 +96,8 @@
                       br_a2)
 
 
+    #TODO: test that fetch correctly does reweaving when needed. RBC 20051008
+
 class TestFetch(TestCaseInTempDir):
 
     def test_fetch(self):

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2005-10-06 08:11:46 +0000
+++ bzrlib/weave.py	2005-10-08 00:03:38 +0000
@@ -74,7 +74,7 @@
 import sha
 from difflib import SequenceMatcher
 
-
+from bzrlib.trace import mutter
 
 
 class WeaveError(Exception):
@@ -83,6 +83,10 @@
 
 class WeaveFormatError(WeaveError):
     """Weave invariant violated"""
+
+
+class WeaveParentMismatch(WeaveError):
+    """Parents are mismatched between two revisions."""
     
 
 class Weave(object):
@@ -398,6 +402,11 @@
         return self._parents[version]
 
 
+    def parent_names(self, version):
+        """Return version names for parents of a version."""
+        return map(self.idx_to_name, self._parents[self.lookup(version)])
+
+
     def minimal_parents(self, version):
         """Find the minimal set of parents for the version."""
         included = self._parents[version]
@@ -774,17 +783,20 @@
         retrieved from self after this call.
 
         It is illegal for the two weaves to contain different values 
-        for any version.
+        or different parents for any version.  See also reweave().
         """
         if other.numversions() == 0:
             return          # nothing to update, easy
+        # two loops so that we do not change ourselves before verifying it
+        # will be ok
         # work through in index order to make sure we get all dependencies
         for other_idx, name in enumerate(other._names):
+            if self._check_version_consistent(other, other_idx, name):
+                continue
+        for other_idx, name in enumerate(other._names):
             # TODO: If all the parents of the other version are already 
             # present then we can avoid some work by just taking the delta
             # and adjusting the offsets.
-            if self._check_version_consistent(other, other_idx, name):
-                continue
             new_parents = self._imported_parents(other, other_idx)
             lines = other.get_lines(other_idx)
             sha1 = other._sha1s[other_idx]
@@ -828,14 +840,98 @@
             n1.sort()
             n2.sort()
             if n1 != n2:
-                raise WeaveError("inconsistent parents for version {%s}: "
-                                 "%s vs %s"
-                                 % (name, n1, n2))
+                raise WeaveParentMismatch("inconsistent parents "
+                    "for version {%s}: %s vs %s" % (name, n1, n2))
             else:
                 return True         # ok!
         else:
             return False
 
+    def reweave(self, other):
+        """Reweave self with other."""
+        new_weave = reweave(self, other)
+        for attr in self.__slots__:
+            setattr(self, attr, getattr(new_weave, attr))
+
+
+def reweave(wa, wb):
+    """Combine two weaves and return the result.
+
+    This works even if a revision R has different parents in 
+    wa and wb.  In the resulting weave all the parents are given.
+
+    This is done by just building up a new weave, maintaining ordering 
+    of the versions in the two inputs.  More efficient approaches
+    might be possible but it should only be necessary to do 
+    this operation rarely, when a new previously ghost version is 
+    inserted.
+    """
+    wr = Weave()
+    ia = ib = 0
+    queue_a = range(wa.numversions())
+    queue_b = range(wb.numversions())
+    # first determine combined parents of all versions
+    # map from version name -> all parent names
+    combined_parents = _reweave_parent_graphs(wa, wb)
+    mutter("combined parents: %r", combined_parents)
+    order = _make_reweave_order(wa._names, wb._names, combined_parents)
+    mutter("order to reweave: %r", order)
+    for name in order:
+        if name in wa._name_map:
+            lines = wa.get_lines(name)
+            if name in wb._name_map:
+                assert lines == wb.get_lines(name)
+        else:
+            lines = wb.get_lines(name)
+        wr.add(name, combined_parents[name], lines)
+    return wr
+
+
+def _reweave_parent_graphs(wa, wb):
+    """Return combined parent ancestry for two weaves.
+    
+    Returned as a list of (version_name, set(parent_names))"""
+    combined = {}
+    for weave in [wa, wb]:
+        for idx, name in enumerate(weave._names):
+            p = combined.setdefault(name, set())
+            p.update(map(weave.idx_to_name, weave._parents[idx]))
+    return combined
+
+
+def _make_reweave_order(wa_order, wb_order, combined_parents):
+    """Return an order to reweave versions respecting parents."""
+    done = set()
+    result = []
+    ia = ib = 0
+    next_a = next_b = None
+    len_a = len(wa_order)
+    len_b = len(wb_order)
+    while ia < len(wa_order) or ib < len(wb_order):
+        if ia < len_a:
+            next_a = wa_order[ia]
+            if next_a in done:
+                ia += 1
+                continue
+            if combined_parents[next_a].issubset(done):
+                ia += 1
+                result.append(next_a)
+                done.add(next_a)
+                continue
+        if ib < len_b:
+            next_b = wb_order[ib]
+            if next_b in done:
+                ib += 1
+                continue
+            elif combined_parents[next_b].issubset(done):
+                ib += 1
+                result.append(next_b)
+                done.add(next_b)
+                continue
+        raise WeaveError("don't know how to reweave at {%s} and {%s}"
+                         % (next_a, next_b))
+    return result
+
 
 def weave_toc(w):
     """Show the weave's table-of-contents"""

